#pragma TextEncoding = "UTF-8"
#pragma rtGlobals=3
#pragma DefaultTab={3,20,4}
#pragma ModuleName=ColorByIndex
#pragma IgorVersion=8
#pragma version=1.21

// Project Updater header
// http://www.igorexchange.com/project/Updater
static constant kProjectID=22305 // the project node on IgorExchange
static strconstant ksShortName="ColorByIndex" // the project short title on IgorExchange

// Set the color of each trace based on a selectable color table and index wave.
// Alternatively, the index value for each trace can be edited in the listbox control.
// A color scale is added to the graph window when the traces are recoloured.

// If a custom color table is selected, when traces are recolored a color
// table wave will be created in the current data folder with a name
// formed by appending "ColorTable" to the graph name. Any existing wave
// with this name will be overwritten.

menu "Graph"
	"Color Traces by Index", /Q,  ColorByIndex#MakePanel()
end

static function/DF getDFR([string package])
	package = SelectString(ParamIsDefault(package), package, "ColorByIndex")
	DFREF dfr = root:Packages:$package
	if (DataFolderRefStatus(dfr) != 1)
		NewDataFolder/O root:Packages
		NewDataFolder/O root:Packages:$package
	endif
	DFREF dfr = root:Packages:$package
	return dfr
end

static function HookCTtraces(STRUCT WMWinHookStruct &s)
	if (s.eventCode == 0) // activate
		wave/T/SDFR=GetDFR() CTtraces
		string strGraph = ParentWindow(s.winName)
		string strList = TraceNameList(strGraph, ";", 5)
		int numTraces = ItemsInList(strList)
		int UpToDate = 1
		int i
		
		if (numTraces != DimSize(CTtraces, 0))
			UpToDate = 0
		else
			for (i=0;i<numTraces;i++)
				if (cmpstr(StringFromList(i, strList), CTtraces[i][0]))
					UpToDate = 0
					break
				endif
			endfor
		endif
		
		if (!UpToDate)
			Duplicate/free/T CTtraces oldCTtraces
			ResetListboxWaves(strGraph)
			int oldrows = DimSize(oldCTtraces, 0)
			int newrows = DimSize(CTtraces, 0)
			if (newrows)
				CTtraces[][1] = ""
				if (oldrows)
					CTtraces[0,(min(oldrows, newrows)-1)][1] = oldCTtraces[p][1]
				endif
			endif
		endif
	elseif (s.eventCode == 2) // kill
		KillWindow/Z $s.winName#G0
		KillDataFolder/Z root:Packages:ColorByIndex
	endif
	return 0
end

static function MakePanel()
	KillWindow/Z CTpanel
	DoUpdate
	string strGraph = WinName(0, 1)
	if (!strlen(strGraph))
		DoAlert 0, "No graph window"
		return 0
	endif
	
	DFREF dfr = getDFR()
	
	NVAR/Z/SDFR=dfr vMin, vMax
	if (NVAR_Exists(vMin)==0 || NVAR_Exists(vMax)==0)
		variable/G dfr:vMin = 0, dfr:vMax = 100
		NVAR/Z/SDFR=dfr vMin, vMax
	endif
	
	Make/T/O/N=(0,2) dfr:CTtraces/wave=CTtraces, dfr:wHelp/wave=wHelp
	Make/O/N=(0,2) dfr:selwave/wave=selwave
	Make/T/O dfr:wTitle = {"trace", "index"}
	
	GetWindow $strGraph wsize
	variable x0 = V_right, y0 = v_top
	NewPanel/HOST=$strGraph/EXT=0/K=1/N=CTpanel/W=(0,0,240,340) as "Color Traces by Index"
	
	SetWindow CTpanel userdata(version) = num2str(GetThisVersion())
	
	PopupMenu popCT, pos={20,10}, size={200,20}, mode=IgorVersion()>=9 ? 61 : 2, value=#"\"*COLORTABLEPOP*\""
	CheckBox chkCustom, pos={20,35}, title="CT wave", value=0, fsize=12, Proc=ColorByIndex#CheckBoxProc
	CheckBox chkCustom, help={"Select a color table wave"}
	PopupMenu popCustom, pos={90,35}, size={150,20}, value=ColorByIndex#ListOfColorTables()
	PopupMenu popCustom, Proc=ColorByIndex#PopupProc, disable=2
	SetVariable svMin, pos={20,65}, size={70,14}, title="Min", fsize=12
	SetVariable svMin, limits={-Inf,Inf,0}, value=dfr:vMin, help={"Index value for color table minimum"}
	SetVariable svMax, pos={130,65}, size={70,14}, title="Max", fsize=12
	SetVariable svMax, limits={-Inf,Inf,0}, value=dfr:vMax, help={"Index value for color table maximum"}
	ListBox list0, pos={20,90}, size={200,150}, listWave=CTtraces, selwave=selwave, userColumnResize=1
	ListBox list0, widths={5,1}, Proc=ColorByIndex#ListboxProc, titleWave=dfr:wTitle, helpWave=wHelp
	GroupBox grpButtons, pos={10,245}, size={220,55}, title="Index values", fsize=12
	Button btnClear, pos={20,270}, size={45,20}, title="Clear", Proc=ColorByIndex#ButtonProc
	Button btnClear, help={"Clear index values"}
	Button btnReset, pos={75,270}, size={45,20}, title="Reset", Proc=ColorByIndex#ButtonProc
	Button btnReset, help={"Resets trace names and sets indices to values spaced evenly\rbetween color table maximum and minimum"}
	Button btnFromWave, pos={130,270}, size={90,20}, title="From wave...", Proc=ColorByIndex#ButtonProc
	Button btnFromWave, help={"Retrieve index values from a numerical wave"}
	Button btnDoIt, pos={170,310}, size={50,20}, title="Do It", Proc=ColorByIndex#ButtonProc
	Button btnDoIt, help={"Reset trace colors and add color scale"}
	
	ResetListboxWaves(strGraph)
	SetWindow $(strGraph)#CTpanel hook(hCTtraces)=ColorByIndex#HookCTtraces
	return 1
end

static function/S ParentWindow(string subwindow)
	return ParseFilePath(0, subwindow, "#", 1, 1)
end

static function ResetListboxWaves(string strGraph)
	DFREF dfr = getDFR()
	wave/T/SDFR=dfr CTtraces, wHelp
	wave/SDFR=dfr selwave
	NVAR/Z/SDFR=dfr vMin, vMax
	
	string strTraces = TraceNameList(strGraph, ";", 5)
	int numTraces = ItemsInList(strTraces)
	Redimension/N=(numTraces,2) CTtraces, selwave, wHelp
	
	if (numTraces)
		CTtraces[][0] = StringFromList(p, strTraces)
		wHelp[][0] = CTtraces // in case we have long trace names
		CTtraces[][1] = num2str(vMin + p/(numTraces-1)*(vMax-vMin))
		selwave = 2 * q // make second column editable
	endif
	return 1
end

static function CheckBoxProc(STRUCT WMCheckBoxAction &s) : CheckboxControl
		
	if (s.eventCode == 2)
		
		if (CheckUpdated(1))
			return 0
		endif
				
		PopupMenu popCT, win=$s.win, disable = 2*s.checked
		PopupMenu popCustom, win=$s.win, disable = 2*(!s.checked)
		KillWindow/Z $s.win#G0
		if (s.checked)
			ControlInfo/W=$s.win popCustom
			wave ctab = CreateCustomColorIndexWave(S_Value, 0)
			wave img = CreateImageWave(ctab)
			Display/W=(20,10,220,30)/HOST=$s.win/N=G0
			AppendImage/T/W=$s.win#G0 img
			ModifyGraph/W=$s.win#G0 margin=-1, frameStyle=1, standoff=0, tick=3, noLabel=2
			SetAxis/A/W=$s.win#G0
			SetWindow $s.win#G0 activeChildFrame=0
			SetActiveSubwindow ##
		endif
	endif
	return 0
end

static function PopupProc(STRUCT WMPopupAction &s) : PopupMenuControl
	if (s.eventCode == 2)

		if (CheckUpdated(0))
			return 0
		endif
				
		string strCT
		if (cmpstr(s.popstr, "Load...") == 0)
			// no way to autoposition a help window :(
			DisplayHelpTopic "Color Table Waves in the Igor Pro Folder"
			PopupMenu $s.ctrlName, win=$s.win, popmatch="none"
			strCT = ""
		else
			strCT = s.popstr
		endif
		
		wave ctab = CreateCustomColorIndexWave(strCT, 0)
		CreateImageWave(ctab)
		
	endif
	return 0
end

static function ListboxProc(STRUCT WMListboxAction &s) : ListBoxControl
	if (s.eventcode == 7)
		s.listWave[s.row][s.col] = ReplaceString("NaN", num2str(str2num(s.listWave[s.row][s.col])), "")
	endif
	return 0
end

static function ButtonProc(STRUCT WMButtonAction &s) : ButtonControl
		
	if (s.eventcode == 2)
		
		if (CheckUpdated(1))
			return 0
		endif
			
		strswitch(s.ctrlName)
			case "btnClear":
				wave/T/SDFR=getDFR() CTtraces
				if (DimSize(CTtraces, 0))
					CTtraces[][1] = ""
				endif
				break
			case "btnDoIt":
				RecolourTraces(s)
				break
			case "btnFromWave":
				ValuesFromWave()
				break
			case "btnReset":
				ResetListboxWaves(ParentWindow(s.win))
				break
		endswitch
	endif
	return 0
end

static function ValuesFromWave()
	wave/T/SDFR=getDFR() CTtraces
	int rows = DimSize(CTtraces, 0)
	if (!rows)
		return 0
	endif
	string strOptions = ""
	sprintf strOptions, "DIMS:1,MINROWS:%d,MAXROWS:%d,TEXT:0", rows, rows
	
	string strList = WaveList("*", ";", strOptions)
	if (!strlen(strList))
		DoAlert 0, "Wave must have " + num2str(rows) + " points, one for each trace"
		return 0
	endif
	
	string strWave = ""
	Prompt strWave, "Select index wave", Popup, strList
	DoPrompt "Select wave containing index values", strWave
	if (v_flag)
		return 0
	endif
	wave wIndex = $strWave
	
	CTtraces[][1] = num2str(wIndex[p])
	CTtraces[][1] = ReplaceString("NaN", CTtraces, "")
	
	return 1
end

static function RecolourTraces(STRUCT WMButtonAction &s)
	ControlInfo/W=CTpanel chkCustom
	int custom = v_value
	string strCT = ""
	
	int error = 0
	DFREF dfr = getDFR()
	NVAR/Z/SDFR=dfr vMin, vMax
	wave/T CTtraces = dfr:CTtraces
	
	if (custom)
		ControlInfo/W=CTpanel popCustom
		if (FindListItem(S_value, "Load...;none;") >=0)
			return 0
		endif
		string strName = CleanupName(ParentWindow(s.win) + "ColorTable", 0)
		Duplicate/O dfr:M_colors $strName/wave = M_colors
	else
		ControlInfo/W=CTpanel popCT
		strCT = S_value
		DFREF dfSav = GetDataFolderDFR()
		SetDataFolder dfr
		ColorTab2Wave $strCT
		wave M_colors
		SetDataFolder dfSav
	endif
			
	variable numColors = DimSize(M_colors, 0)
	int i
	string strTrace
	variable pnt, val
	
	// colour the traces
	for (i=DimSize(CTtraces, 0)-1;i>=0;i--)
		strTrace = CTtraces[i][0]
		val = str2num(CTtraces[i][1])
		if (val != limit(val, min(vMin,vMax), max(vMin,vMax)))
			error = error | 1
		endif
		
		pnt = floor((str2num(CTtraces[i][1])-vMin)/(vMax-vMin)*(numColors))
		pnt = limit(pnt, 0, numColors-1)
		if (numtype(pnt))
			error = error | 2
			continue
		endif
		
		ModifyGraph rgb($strTrace) = (M_colors[pnt][0], M_colors[pnt][1], M_colors[pnt][2])
		
	endfor
	
	// draw the colour scale
	if (custom)
		SetScale/I x, vMin, vMax, M_colors
		ColorScale/C/N=ct/F=0/A=RC cindex=M_colors
	else
		ColorScale/C/N=ct/F=0/A=RC ctab={vMin,vMax,$strCT,0}
	endif
	
	if (error)
		DoAlert 0, "Index of at least one trace is out of range.\rColors of out-of-range traces are not changed."
	endif
end

// create an image wave to display color table in panel
static function/wave  CreateImageWave(wave ctab)
	DFREF dfr = getDFR()
	Make/O/N=(DimSize(ctab,0), 2, 3)/W/U dfr:CustomImg/wave=CustomImg
	CustomImg = ctab[p][r]
	return CustomImg
end

// create a color table index wave in the package folder
static function/wave CreateCustomColorIndexWave(string strCT, int reversed)
	DFREF dfr = getDFR()
	DFREF tableDFR = getDFR(package = "ColorTables")
	
	if (strlen(strCT))
		if (IgorVersion() >= 9)
			wave/Z w = FindWaveRecursive(strCT, tableDFR)
		else
			wave/Z w = $strCT
		endif
	endif
	if (!WaveExists(w))
		// make a dummy color table
		Make/O/U/I/N=(2,3) tableDFR:none = 60000
		wave w = tableDFR:none
	endif
	
	NVAR/Z/SDFR=dfr vMin, vMax
	Duplicate/O w dfr:M_colors/wave=M_colors
	SetScale/I x, vMin, vMax, M_colors

	return M_colors
end

static function/wave FindWaveRecursive(string strCT, DFREF dfr)
	int i
	for (i=0;1;i++)
		wave/Z w = WaveRefIndexedDFR(dfr, i)
		if (!WaveExists(w))
			break	// no more waves
		endif
		// see if it is our color table wave
		if (cmpstr(NameOfWave(w), strCT)==0)
			return w
		endif
	endfor
	
	// recurse into all subfolders
	variable numChildDataFolders = CountObjectsDFR(dfr, 4)
	for (i=0; i<numChildDataFolders; i+=1)
		DFREF childDFR = dfr:$GetIndexedObjNameDFR(dfr, 4, i)
		wave/Z w = FindWaveRecursive(strCT, childDFR)
		if (WaveExists(w))
			return w
		endif
	endfor
	
	return $""
end

// following three functions adapted from KBColorizeTraces.ipf
static function/S ListOfColorTables()
	if (IgorVersion() < 9)
		return RecurseColorTablesFolder("", GetDataFolderDFR(), recurse=0)
	endif
	return RecurseColorTablesFolder("", root:Packages:ColorTables) + "Load...;"
end

// returns 1 if w is a color table wave (numeric, 3 or 4 columns, more than 1 row)
static function IsColorTableWave(WAVE w)

	if (WaveType(w, 1) != 1)
		return 0	// not numeric!
	endif
	
	int cols = DimSize(w, 1)
	if (cols < 3 || cols > 4)
		return 0
	endif
	
	if (DimSize(w, 0) < 2) // not enough rows
		return 0
	endif
	
	return 1
end

static function/S RecurseColorTablesFolder(string listSoFar, DFREF dfr, [int recurse])

	if (!DataFolderRefStatus(dfr))
		return listSoFar
	endif
	
	recurse = ParamIsDefault(recurse) ? 1 : recurse
	
	// iterate through waves in folder
	int i
	for (i=0;1;i++)
		wave/Z w = WaveRefIndexedDFR(dfr, i)
		if (!WaveExists(w))
			break // no more waves
		endif
		// see if it is a color table wave
		if (IsColorTableWave(w))
			listSoFar += NameOfWave(w) + ";"
		endif
	endfor
	
	// then recurse into all subfolders
	if (recurse)
		variable numChildDataFolders = CountObjectsDFR(dfr, 4)
		for (i=0; i<numChildDataFolders; i+=1)
			DFREF childDFR = dfr:$GetIndexedObjNameDFR(dfr, 4, i)
			listSoFar = RecurseColorTablesFolder(listSoFar, childDFR)
		endfor
	endif
	
	return listSoFar
end

// returns truth that this procedure file has been updated since initialisation
static function CheckUpdated(int restart)
	int version = GetThisVersion()
	int panelVersion = str2num(GetUserData("CTpanel", "", "version"))
	if (panelVersion != version)
		if (restart)
			DoAlert 0, "You have updated the package since this panel was created.\r\rThe package will restart to update the control panel."
			MakePanel()
		else
			DoAlert 0, "You have updated the package since this panel was created.\r\rPlease close and reopen the panel to continue."
		endif
		return 1
	endif
	return 0
end

// ProcedureVersion("") doesn't work for independent modules!
// extract procedure version from this file
static function GetThisVersion()

	variable procVersion
	// try the quick way
	#if exists("ProcedureVersion") == 3
	procVersion = ProcedureVersion("")
	if (procVersion)
		return 100 * procVersion
	endif
	#endif
	
	int maxLines = 20 // number of lines to search for version pragma
	int refNum, i
	string strHeader = ""
	string strLine = ""
	
	Open/R/Z refnum as FunctionPath("")
	if (refnum == 0)
		return 0
	endif
	for (i=0;i<maxLines;i+=1)
		FReadLine refNum, strLine
		strHeader += strLine
	endfor
	Close refnum
	wave/T ProcText = ListToTextWave(strHeader, "\r")
	Grep/Q/E="(?i)^#pragma[\s]*version[\s]*=" /LIST/Z ProcText
	if (v_flag != 0)
		return 0
	endif
	s_value = LowerStr(TrimString(s_value, 1))
	sscanf s_value, "#pragma version = %f", procVersion
	ProcVersion = (V_flag!=1 || ProcVersion<=0) ? 0 : ProcVersion

	return 100 * ProcVersion
end